#include <fstream>
#include <sstream>
#include <iostream>
#include "hair.h"

// some helper function
inline unsigned int readUInt(std::ifstream &fs) {
	unsigned int temp;
	fs.read(reinterpret_cast<char *>(&temp), sizeof(unsigned int));
	return temp;
}
inline float readSingle(std::ifstream &fs) {
	float temp;
	fs.read(reinterpret_cast<char *>(&temp), sizeof(float));
	return temp;
}
inline void writeUInt(std::ofstream &fs, unsigned int val) {
	fs.write(reinterpret_cast<char *>(&val), sizeof(unsigned int));
}
inline void writeSingle(std::ofstream &fs, float val) {
	fs.write(reinterpret_cast<char *>(&val), sizeof(float));
}

bool importHairBinary(const std::string &path, Hair& hair, bool hasUV = false) {
	std::ifstream inFile(path, std::ios::binary | std::ios::in);

	/* mitsuba uses a 11 bytes header to idendify binary format. */
	inFile.seekg(11);
	if(inFile.bad()) return false;

	size_t vertexCount = readUInt(inFile);
	std::cout << "Loading " << vertexCount << " hair vertices ..\n";

	Hair _hair;

	auto &vertices = _hair.vertices;
	auto &vertexStartsFiber = _hair.vertexStartsFiber;
	auto &verticesTexCoord = _hair.verticesTexCoord;

	vertices.reserve(vertexCount);
	vertexStartsFiber.reserve(vertexCount);
	verticesTexCoord.reserve(vertexCount);
	hair.hairCount = 0;

	bool newFiber = true;
	size_t verticesRead = 0;
	Point3 p;
	Point2 uv;

	while(verticesRead != vertexCount) {
		if(inFile.bad()) 
			return false;
		Float value = readSingle(inFile);
		if(std::isinf(value)) {
			if(hasUV) {
				uv.x = readSingle(inFile);
				uv.y = readSingle(inFile);
			}
			p.x = readSingle(inFile);
			p.y = readSingle(inFile);
			p.z = readSingle(inFile);
			newFiber = true;
			_hair.hairCount++;
		}
		else {
			p.x = value;
			p.y = readSingle(inFile);
			p.z = readSingle(inFile);
		}
		verticesRead++;

		vertices.push_back(p);
		vertexStartsFiber.push_back(newFiber);
		verticesTexCoord.push_back(uv);
		newFiber = false;
	}
	hair = _hair;
	return true;
}

bool importHairText(const std::string &path, Hair &hair) {
	std::ifstream inFile(path, std::ios::binary | std::ios::in);

	Hair _hair;
	auto &vertices = _hair.vertices;
	auto &vertexStartsFiber = _hair.vertexStartsFiber;
	auto &verticesTexCoord = _hair.verticesTexCoord;
	_hair.hairCount = 0;

	bool newFiber = true;
	Point2 uv;
	Point3 p;
	std::string line;

	while(inFile.good()) {
		std::getline(inFile, line);
		if(line.length() == 0 || line.length() > 0 && line[0] == '#') {
			newFiber = true;
			continue;
		}
		if(newFiber) {
			// Read uv when a new fiber is encountered
			std::istringstream iss(line);
			iss >> uv.x >> uv.y;
			std::getline(inFile, line);
			_hair.hairCount++;
		}
		std::istringstream iss(line);
		iss >> p.x >> p.y >> p.z;
		if(!iss.fail()) {
			vertices.push_back(p);
			vertexStartsFiber.push_back(newFiber);
			verticesTexCoord.push_back(uv);
			newFiber = false;
		}
		else {
			newFiber = true;
		}
	}
	hair = _hair;
	return true;
}

bool importHair(const std::string &path, Hair& hair) {
	char temp[11];
	std::ifstream in(path, std::ios::binary | std::ios::in);
	in.read(temp, 11);
	/* default binary hair file in mitsuba does not contains uv,
		our format should use a different header here to distinguish. 
		however we assume all input binary hair files come with uv.*/

	//if(memcmp(temp, "BINARY_HAIR", 11) == 0) {
	//	return importHairBinary(path, hair, false);
	//}
	//else if(memcmp(temp, "WITHUV_HAIR", 11) == 0) {
	//	return importHairBinary(path, hair, true);
	//}
	if(memcmp(temp, "BINARY_HAIR", 11) == 0) {
		return importHairBinary(path, hair, true);
	}
	return importHairText(path, hair);
}

bool exportHairText(const std::string &path, const Hair &hair) {
	std::ofstream outFile(path, std::ios::binary | std::ios::out | std::ios::trunc);
	if(!outFile.is_open()) {
		std::cerr << "Unable to create output hair file.\n";
		return false;
	}
	auto &vertices = hair.vertices;
	auto &vertexStartsFiber = hair.vertexStartsFiber;
	auto &verticesTexCoord = hair.verticesTexCoord;

	for(size_t i = 0; i < vertices.size(); ++i)
	{
		if(vertexStartsFiber[i]) {
			outFile << verticesTexCoord[i].x << ' ' << verticesTexCoord[i].y << '\n'; \
		}
		outFile << vertices[i].x << ' '
			<< vertices[i].y << ' '
			<< vertices[i].z << '\n';
		if(i + 1 < vertices.size() && vertexStartsFiber[i + 1])
			outFile << '\n';
	}
	return true;
}

bool exportHairBinary(const std::string &path, const Hair &hair) {
	std::ofstream outFile(path, std::ios::binary | std::ios::out | std::ios::trunc);
	if(!outFile.is_open()) { 
		std::cerr << "Unable to create output hair file.\n";
		return false;
	}
	auto &vertices = hair.vertices;
	auto &vertexStartsFiber = hair.vertexStartsFiber;
	auto &verticesTexCoord = hair.verticesTexCoord;

	/* should write a different header here, however we assume 
		all binary hair files come with uv in mitsuba side.*/

	//outFile.write("WITHUV_HAIR", 11);
	outFile.write("BINARY_HAIR", 11);
	writeUInt(outFile, (unsigned int) vertices.size());
	constexpr float INF = std::numeric_limits<float>::infinity();
	for(size_t i = 0; i < vertices.size(); ++i)
	{
		if(vertexStartsFiber[i]) {
			writeSingle(outFile, INF);
			writeSingle(outFile, (float) verticesTexCoord[i].x);
			writeSingle(outFile, (float) verticesTexCoord[i].y);
		}
		writeSingle(outFile, (float) vertices[i].x);
		writeSingle(outFile, (float) vertices[i].y);
		writeSingle(outFile, (float) vertices[i].z);
	}
	return true;
}


bool exportInfo(std::string path, const std::vector<Clump> &clumpInfo) {
	std::ofstream outFile(path, std::ios::out | std::ios::binary | std::ios::trunc);
	if(!outFile.is_open()) {
		std::cerr << "Unable to create output info file.\n";
		return false;
	}
	constexpr char *HEADER = "BINARY_INFO";
	constexpr float INF = std::numeric_limits<float>::max();
	outFile.write(HEADER, 11);
	writeUInt(outFile, (unsigned int) clumpInfo.size());
	for(const auto &info : clumpInfo) {
		writeSingle(outFile, (float) info.clumpRadius);
		writeSingle(outFile, (float) info.texBoundU.x);
		writeSingle(outFile, (float) info.texBoundV.x);
		writeSingle(outFile, (float) info.texBoundU.y);
		writeSingle(outFile, (float) info.texBoundV.y);
	}
	return true;
}